#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lcd.h"

#define	width	14
#define	height	6

#define	LEFT	1
#define	RIGHT	2
#define	DOWN	3
#define	UP		4
#define NONE	5

typedef struct coord {
	int x;
	int y;
} coord;

static int moveDirection;
static int pressedButton;
static coord snake[width * height];
static coord food;
static coord bonus;
static int size;
static int score;
static int bonusFrames;
static int bonusActivated;
static int countFoodToBonus;

/* Setare directii porturi folosite */
void init_buttons(void) {

    DDRD = 0xff;
    PORTD = 0x00;

    // LEFT
    DDRD &= ~(1 << PD0);
    PORTD |= (1 << PD0);
    
    // DOWN
    DDRD &= ~(1 << PD1);
    PORTD |= (1 << PD1);
    
    // UP
    DDRD &= ~(1 << PD5);
    PORTD |= (1 << PD5);
    
    // RIGHT
    DDRD &= ~(1 << PD4);
    PORTD |= (1 << PD4);
}

/* Verificare butoane apasate */
void check_input(void) {

    // LEFT
    if (!(PIND & (1 << PD0)) && moveDirection != RIGHT && moveDirection != LEFT) {
        pressedButton = LEFT;
		_delay_ms(100);
    }
    
    // DOWN
    if (!(PIND & (1 << PD1)) && moveDirection != UP && moveDirection != DOWN) {
        pressedButton = DOWN;
		_delay_ms(100);
    }
    
    // UP
    if (!(PIND & (1 << PD5)) && moveDirection != DOWN && moveDirection != UP) {
        pressedButton = UP;
		_delay_ms(100);
    }
    
    // RIGHT
    if (!(PIND & (1 << PD4)) && moveDirection != LEFT && moveDirection != RIGHT) {
        pressedButton = RIGHT;
		_delay_ms(100);
    }
}

void init_snake() {
	int i;

	moveDirection = LEFT;
	pressedButton = NONE;
	score = 0;
	bonusFrames = 25;
	bonusActivated = 0;
	countFoodToBonus = 0;

	size = 3;

	for (i = 0; i < size; ++i) {
		snake[i].y = 3;
		snake[i].x = 6 + i;
	}
}

void generate_foodAndBonus(coord *aux) {
	int i, check = 0;
	srand(size * rand());
	
	while (!check) {
		aux->x = rand() % width + 1;
		aux->y = rand() % height + 1;
		
		for (i = 0; i < size; ++i) {
			if (snake[i].x == aux->x && snake[i].y == aux->y) {
				check = 1;
				break;
			}
		}
		
		if (check)
			check = 0;
		else
			check = 1;
	}
}

void check_collision_with_food() {

	if (snake[0].x == food.x && snake[0].y == food.y) {
		int x = snake[size - 1].x - snake[size - 2].x;
		int y = snake[size - 1].y - snake[size - 2].y;
		snake[size].x = snake[size - 1].x + x;
		snake[size].y = snake[size - 1].y + y;
		size++;
		score += 5;
		countFoodToBonus++;
		
		if (countFoodToBonus % 5 == 0) {
			bonusActivated = 1;
			generate_foodAndBonus(&bonus);
		}
			
		generate_foodAndBonus(&food);
	}
}

void check_collision_with_bonus() {
	int i;

	if (bonusActivated) {
		for (i = 0; i < size; ++i) {
			if (snake[i].x == bonus.x && snake[i].y == bonus.y) {
				score += bonusFrames * 3;
				bonusActivated = 0;
				break;
			}
		}
		
		if (bonusActivated) {
			bonusFrames--;
			if (bonusFrames == 0)
				bonusActivated = 0;
		}
	}

}

void check_collision_with_snake() {
	int i;
	
	for (i = 1; i < size; ++i) {
		if (snake[0].x == snake[i].x && snake[0].y == snake[i].y) {
			int j;
			char *points = (char*)malloc(sizeof(char) * 3);
			itoa(score, points, 10);
			lcd_clear();
			lcd_goto_xy(3, 3);
			for (j = 0; j< size; j++) {
				snake[j].x = 0;
				snake[j].y = 0;
			}
			
			lcd_str("GAME OVER");
			lcd_goto_xy(7, 4);
			lcd_str(points);
			free(points);
			init_snake();
			_delay_ms(5000);
			break;
		}
	}

}


void draw_frame() {
	int i;
	
	if (bonusActivated) {
		lcd_goto_xy(bonus.x, bonus.y);
		lcd_chr('O');
	}
	
	lcd_goto_xy(food.x, food.y);
	lcd_chr('X');
	
	lcd_goto_xy(snake[0].x, snake[0].y);
	
	if (moveDirection == LEFT)
		lcd_chr('<');
	else if (moveDirection == RIGHT)
		lcd_chr('>');
	else if (moveDirection == UP)
		lcd_chr('^');
	else
		lcd_chr('v');

	for (i = 1; i < size; ++i) {
		lcd_goto_xy(snake[i].x, snake[i].y);
		lcd_chr('#');
	}
}

void move_snake() {
	int i;
	
	for (i = size - 1; i > 0; --i) {
		snake[i].x = snake[i - 1].x;
		snake[i].y = snake[i - 1].y;
	}

	if (moveDirection == LEFT || moveDirection == RIGHT) {

		if (pressedButton == UP) {
			snake[0].y = (snake[0].y - 1) ? (snake[0].y - 1) : height;
			moveDirection = UP;
		} else if (pressedButton == DOWN) {
			snake[0].y = snake[0].y % height + 1;
			moveDirection = DOWN;
		} else if (pressedButton == NONE && moveDirection == RIGHT) {
			snake[0].x = snake[0].x % width + 1;
		} else if (pressedButton == NONE && moveDirection == LEFT) {
			snake[0].x = (snake[0].x - 1 ) ? (snake[0].x - 1) : width;
		}
	} else if (moveDirection == UP || moveDirection == DOWN) {

		if (pressedButton == LEFT) {
			snake[0].x = (snake[0].x - 1) ? (snake[0].x - 1) : width;
			moveDirection = LEFT;
		} else if (pressedButton == RIGHT) {
			snake[0].x = snake[0].x % width + 1;
			moveDirection = RIGHT;
		} else if (pressedButton == NONE && moveDirection == DOWN) {
			snake[0].y = snake[0].y % height + 1;
		} else if (pressedButton == NONE && moveDirection == UP) {
			snake[0].y = (snake[0].y - 1) ? (snake[0].y - 1) : height;
		}
	}
}

int main(void)
{
	// Setup LCD
	lcd_init();
	lcd_contrast(0x40);
	init_buttons();

	init_snake();
	generate_foodAndBonus(&food);

	while(1) {
		draw_frame();
		check_input();
		_delay_ms(150);
		lcd_clear();
		move_snake();
		check_collision_with_food();
		check_collision_with_bonus();
		check_collision_with_snake();
		pressedButton = NONE;
	}
}

